import base64
import hashlib
import threading
import urllib.parse
from http.server import BaseHTTPRequestHandler, HTTPServer

AUTH_STATE = {
    "received_callback": False,
    "code": None,
    "error_message": None,
    "received_state": None,
}


class SimpleServer(BaseHTTPRequestHandler):
    def _set_headers(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def _html(self, message):
        content = f"<html><body><h1>{message}</h1></body></html>"
        return content.encode("utf8")

    def do_GET(self):
        if "/callback?" not in self.path:
            return

        query = urllib.parse.urlparse(self.path).query
        params = urllib.parse.parse_qs(query)

        self._set_headers()
        self.wfile.write(self._html("Please return to your application now."))

        AUTH_STATE["error_message"] = params.get("error", [""])[0]
        AUTH_STATE["code"] = params.get("code", [""])[0]
        AUTH_STATE["received_state"] = params.get("state", [""])[0]
        AUTH_STATE["received_callback"] = True


class ServerThread(threading.Thread):
    """
    The simple server is done this way to allow shutting down after a single request has been received.
    """

    def __init__(self):
        threading.Thread.__init__(self)
        self.srv = HTTPServer(("127.0.0.1", 64242), SimpleServer)

    def run(self):
        self.srv.serve_forever()

    def shutdown(self):
        self.srv.shutdown()


def auth0_url_encode(byte_data):
    """
    Safe encoding handles + and /, and also replace = with nothing
    :param byte_data:
    :return:
    """
    return base64.urlsafe_b64encode(byte_data).decode("utf-8").replace("=", "")


def generate_challenge(a_verifier):
    return auth0_url_encode(hashlib.sha256(a_verifier.encode()).digest())



import os

from whiffle_client.common.mapping.base import BaseMapping
from whiffle_client.io import stream_download_write_chunks


VALID_FORMATS = ["nc", "csv"]


class AtlasEndpoints(BaseMapping):
    URL = "/api/v1/atlas"

    def download_point(
        self,
        longitude: float,
        latitude: float,
        format: str = "nc",
        dataset: str = "meso",
        output_name: str = None,
        output_dir: str = None,
    ):
        """Download atlas data

        Parameters
        ----------
        longitude : float
            Longitude location of data to download
        latitude : float
            Latitude location of data to download
        format : str, optional
            Format in which the data will be downloaded
        dataset : str, optional
            Requested dataset for data download
        output_name : str, optional
            Output file name to use if provided, else use same name as remote file, by default None
        output_dir : str, optional
            Output directory path, by default None

        Raises
        ------
        ValueError
            Raises error if file format is not in the valid file formats
        """

        if format not in VALID_FORMATS:
            raise ValueError(
                f"Atlas download request has to be one of the valid formats: {VALID_FORMATS}"
            )

        download_url = f"{self.session.server_url}{self.URL}/time-series.{format}?latitude={latitude}&longitude={longitude}"
        download_response = self.session.get_request(download_url, stream=True)

        # NOTE: If provided, use given name, else take same name as response filename or return default
        local_filename = (
            output_name
            or download_response.headers.get(
                "Content-Disposition", f"filename=atlas_data.{format}"
            ).split("filename=")[1]
        )
        output_dir = output_dir or os.getcwd()

        print(
            f"Fetch atlas:{dataset} data at location "
            f"longitude:{longitude} latitude:{latitude} and "
            f"store it at: `{output_dir}/{local_filename}`"
        )
        stream_download_write_chunks(
            f"{output_dir}/{local_filename}", download_response
        )

    download = download_point



from pathlib import Path
from typing import Union

from whiffle_client.forecast.components.asset_model import AssetModel
from whiffle_client.decorators import load_data
from whiffle_client.common.mapping.base import BaseMapping


class AssetEndpoints(BaseMapping):
    """Asset model endpoint

    Parameters
    ----------
    BaseMapping : Basemapping
        Base mapping
    """

    URL = "/api/v1/assets"
    RESOURCE_TYPE = "Asset"

    # pylint: disable=arguments-differ
    @load_data(AssetModel)
    def add(
        self,
        data: Union[str, dict, Path, AssetModel] = None,
    ) -> AssetModel:
        """Add new Asset

        Parameters
        ----------
        data : Union[str, dict, Path, AssetModel], optional
            Either a path to a yaml, the data itself as a dictionary or a AssetModel object
            containing the parameters that define the Asset, by default None

        Returns
        -------
        AssetModel
            Object instance of the wind simulation
        """
        asset = self.session.post_request(
            f"{self.session.server_url}{self.URL}", data=data
        )
        return AssetModel.from_dict(asset.json())

    def get(self, asset_name: str) -> AssetModel:
        """Get a Asset

        Parameters
        ----------
        asset_name : str
            Name of the asset

        Returns
        -------
        AssetModel
            Requested AssetModel
        """
        return AssetModel.from_dict(
            self.session.get_request(
                f"{self.session.server_url}{self.URL}/{asset_name}",
            ).json()
        )

    # pylint: disable=arguments-differ
    def get_all(self) -> list[AssetModel]:
        """Get a list of all Assets available to the user

        Returns
        -------
        list[AssetModel]
            List of AssetModels object instances
        """
        assets = []
        request = self.session.get_request(f"{self.session.server_url}{self.URL}")
        data = request.json()
        assets.extend(AssetModel.from_dict(asset) for asset in data.get("results", []))
        while isinstance(data.get("next"), str):
            data = self.session.get_request(data["next"]).json()
            assets.extend(
                AssetModel.from_dict(asset) for asset in data.get("results", [])
            )
        return assets

    def edit(self):
        """Abstract method for edit

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def delete(self):
        """Abstract method for delete

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def download(self):
        """Abstract method for download

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError



import datetime
from dataclasses import dataclass, field

from whiffle_client.common.components.base import BaseComponent


# pylint: disable=duplicate-code
@dataclass
class ForecastModel(BaseComponent):
    """Forecast model"""

    reference_time: datetime.datetime = field(default=None)
    asset_name: int = field(default=None)
    name: str = field(default="blender")
    steps: dict = field(default=None)



import numpy as np
import matplotlib as mpl


def plot_one_to_one_line(ax, x, y, method, **kw):
    if method in [None, "max"]:
        start = min(np.nanmin(x), np.nanmin(y))
        end = max(np.nanmax(x), np.nanmax(y))
    elif method == "percentile":
        percmin, percmax = 1, 99
        start = min(np.nanpercentile(x, percmin), np.nanpercentile(y, percmin))
        end = max(np.nanpercentile(x, percmax), np.nanpercentile(y, percmax))
    elif method == "margin":
        margin = 0.1
        start = (1 - margin) * min(np.nanmin(x), np.nanmin(y))
        end = (1 + margin) * max(np.nanmax(x), np.nanmax(y))
    else:
        raise ValueError(method)

    ax.plot([start, end], [start, end], "-", **kw)

    ax.set_xlim(start, end)
    ax.set_ylim(start, end)

    ax.set_aspect(1)


def scatter_density(
    ax,
    x,
    y,
    cmap="viridis",
    one_to_one=None,
    s=1,
    edgecolor="none",
    sort=False,
    colour_kwargs={},
    **kw,
):
    """Efficient scatter density algorithm."""

    x, y = x.to_numpy().reshape(-1, 1), y.to_numpy().reshape(-1, 1)

    inds = np.isfinite(x) & np.isfinite(y)

    x = x[inds]
    y = y[inds]

    if sort or colour_kwargs:
        x, y, cvals = hist_colouring_sort(x, y, **colour_kwargs)
        C = ax.scatter(x, y, c=cvals, cmap=cmap, edgecolor=edgecolor, s=s, **kw)
    else:
        cvals = hist_colouring(x, y, 1000, cmap, 15)
        C = ax.scatter(x, y, c=cvals, edgecolor=edgecolor, s=s, **kw)

    if one_to_one is not False:
        plot_one_to_one_line(ax, x, y, method=one_to_one, color="k")

    return C


def hist_colouring(
    xarray, yarray, nbins=300, cmap="inferno", smoother=3, fnc=None, norm=None
):
    """
    Explanation on `norm`:

    Purpose:
      To have multiple figures showing data of different min/maxes (thus dif-
      ferent bin-edges) and different "maximum bin value"s, show comparable
      color behavior for comparable point-densities.

    Implications:
      Color behavior normalization is by default determined by the maximum bin-
      count. This bin (or smoothed bin) will have the highest value color.
      The point-density is the number of points per area. It is therefore
      necessary to also consider the area of the bins in the normalization.
    """
    cmap = mpl.cm.get_cmap(cmap)

    def bin_indices(data, bins):
        ixs = np.zeros(len(data), int) - 1
        for i, bini in enumerate(bins[:-1]):
            inbin = (bini <= data) * (data < bins[i + 1])
            if i == len(bins) - 1:
                inbin *= data == bins[i + 1]
            ixs += inbin * (i + 1)
        return ixs

    mask = np.isfinite(xarray) & np.isfinite(yarray)
    xarray = xarray[mask]
    yarray = yarray[mask]

    xmin = np.min(xarray)
    xmax = np.max(xarray)
    ymin = np.min(yarray)
    ymax = np.max(yarray)

    xbins = np.linspace(
        xmin - (xmax - xmin) / (nbins * 10),
        xmax + (xmax - xmin) / (nbins * 10),
        nbins + 1,
    )
    ybins = np.linspace(
        ymin - (ymax - ymin) / (nbins * 10),
        ymax + (ymax - ymin) / (nbins * 10),
        nbins + 1,
    )

    hist2d, _, _ = np.histogram2d(xarray, yarray, bins=[xbins, ybins])

    if fnc is not None:
        hist2d = fnc(hist2d)
    x_ix = bin_indices(xarray, xbins)
    y_ix = bin_indices(yarray, ybins)

    if smoother:
        smoothhist = np.zeros(len(x_ix))
        for i, (ix, iy) in enumerate(zip(x_ix, y_ix)):
            smoothhist[i] = np.mean(
                hist2d[
                    max(0, ix - smoother) : ix + smoother + 1,
                    max(0, iy - smoother) : iy + smoother + 1,
                ]
            )

        if norm is None:
            norm = np.max(smoothhist)
        else:
            dA = (xbins[1] - xbins[0]) * (ybins[1] - ybins[0]) * nbins**2
            norm *= dA

        return cmap(smoothhist / norm)

    else:
        if norm is None:
            norm = np.max(hist2d)
        else:
            dA = (xbins[1] - xbins[0]) * (ybins[1] - ybins[0]) * nbins**2
            norm *= dA

        return cmap(hist2d[x_ix, y_ix] / norm)


def hist_colouring_sort(x, y, nbins=300, zmax=None, fnc=None):
    """
    Alternative to hist_colouring that uses sorting instead of smoothing
    and allows for a colorbar to be added easily
    """
    if fnc == "log":
        fnc = lambda z: np.log10(z, where=z > 0)

    hist2d, locx, locy = np.histogram2d(x, y, bins=[nbins, nbins])

    if zmax:
        hist2d = hist2d.clip(0, zmax)

    if fnc:
        hist2d = fnc(hist2d)

    # Sorting the points such that that the largest bins are drawn on top
    hist2d = np.array(
        [
            hist2d[np.argmax(a <= locx[1:]), np.argmax(b <= locy[1:])]
            for a, b in zip(x, y)
        ]
    )
    idx = hist2d.argsort()
    x, y, hist2d = x[idx], y[idx], hist2d[idx]

    return x, y, hist2d



from abc import ABC


class BaseMapping(ABC):
    """Base mapping

    Parameters
    ----------
    ABC : AbstractBaseClass
        Abstract base class

    Raises
    ------
    NotImplementedError
        add not implemented
    NotImplementedError
        edit not implemented
    NotImplementedError
        get_all not implemented
    NotImplementedError
        get not implemented
    NotImplementedError
        delete not implemented
    NotImplementedError
        download not implemented
    """

    URL = None
    RESOURCE_TYPE = None

    def __init__(self, session):
        self.session = session

    def add(self):
        """Abstract method for add

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def edit(self):
        """Abstract method for edit

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def get_all(self):
        """Abstract method for get_all

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def get(self):
        """Abstract method for get

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def delete(self):
        """Abstract method for delete

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def download(self):
        """Abstract method for download

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError



import numpy as np
import pandas as pd
import xarray as xr

from whiffle_client.analysis.utils import mean_angle

WINDOGRAPHER_COLUMN_MAP = {
    "AvgWs": "horizontal-wind-speed [m/s]",
    "AvgWd": "wind-direction [deg]",
}

CABAUW_DATA_VARS_MAP = {
    "F": ("horizontal-wind-speed", "m/s"),
    "D": ("wind-direction", "deg"),
}
CABAUW_DIMS_MAP = {
    "z": ("height", "m"),
}
CABAUW_HEIGHTS_TO_KEEP = [10.0, 20.0, 40.0, 80.0, 140.0, 200.0]
WHIFFLE_WIND_BACKWARD_COMPATIBILITY_MAP = {
    "wind-speed [m/s]": "horizontal-wind-speed [m/s]",
}
WHIFFLE_WIND_METMAST_SINGLE_HEIGHT_VARS = ["atmospheric-boundary-layer-height"]


def _parse_turbine_column_name(col_name: str) -> tuple[str, str, str]:
    turbine_name = "_".join(col_name.split("_")[:-1])
    variable = col_name.split("_")[-1].split("[")[0]
    units = col_name.split("[")[-1].split("]")[0]
    units = f"[{units}]"
    return turbine_name, variable, units


def _parse_metmast_multi_height_column_name(
    col_name: str,
) -> tuple[str, float, str, str]:
    metmast_name = "_".join(col_name.split("_")[:-2])
    height = float(col_name.split("_")[-2][1:])
    variable = col_name.split("_")[-1].split("[")[0]
    units = col_name.split("[")[-1].split("]")[0]
    units = f"[{units}]"
    return metmast_name, height, variable, units


def _parse_metmast_single_height_column_name(
    col_name: str,
) -> tuple[str, float, str, str]:
    metmast_name = "_".join(col_name.split("_")[:-1])
    variable = col_name.split("_")[-1].split("[")[0]
    units = col_name.split("[")[-1].split("]")[0]
    units = f"[{units}]"
    return metmast_name, np.nan, variable, units


def _read_whiffle_wind_csv(csv_path):
    df = pd.read_csv(csv_path, sep=";")
    df["time[UTC]"] = pd.to_datetime(df["time[UTC]"], utc=True)
    df = df.set_index("time[UTC]").rename_axis("time [UTC]")
    return df


def _parse_windographer_column_name_to_tuple(col_name):
    var, height, boom_angle = col_name.split()[1:4]
    height = np.floor(float(height.split("m")[0]))
    return var, height, boom_angle


def _split_metmast_columns(columns):
    multi_height_cols = [
        col
        for col in columns
        for var in WHIFFLE_WIND_METMAST_SINGLE_HEIGHT_VARS
        if var not in col
    ]
    single_height_cols = list(set(columns) - set(multi_height_cols))
    return multi_height_cols, single_height_cols


def whiffle_wind_turbines_csv_to_dataframe(csv_path: str) -> pd.DataFrame:
    """Load a whiffle wind csv with turbine data into a pandas dataframe.

    Parameters
    ----------
    csv_path : str
        Path to the whiffle wind csv file with turbine data.

    Returns
    -------
    pd.DataFrame
        Dataframe with simulated wind turbine data.
    """
    df = _read_whiffle_wind_csv(csv_path)
    midx = df.columns.map(_parse_turbine_column_name)
    midx = [(x[0], f"{x[1]} {x[2]}") for x in midx]
    midx = pd.MultiIndex.from_tuples(midx, names=("turbine", None))
    df.columns = midx
    df = df.stack(level=0, future_stack=True).sort_index()
    df = df.rename(columns=WHIFFLE_WIND_BACKWARD_COMPATIBILITY_MAP)
    return df


def whiffle_wind_turbines_csv_to_dataset(csv_path: str) -> xr.Dataset:
    """Load a whiffle wind csv with turbine data into a xarray dataset.

    Parameters
    ----------
    csv_path : str
        Path to the whiffle wind csv file with turbine data.

    Returns
    -------
    xr.Dataset
        Dataset with simulated wind turbine data.
    """
    df = whiffle_wind_turbines_csv_to_dataframe(csv_path)
    ds = turbine_dataframe_to_dataset(df)
    return ds


def turbine_dataframe_to_dataset(df: pd.DataFrame) -> xr.Dataset:
    """Convert a dataframe with turbine data into an xarray dataset.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe with simulated wind turbine data.

    Returns
    -------
    xr.Dataset
        Dataset with simulated wind turbine data.
    """
    df = df.rename_axis(("time", "turbine"))
    parsed_columns = [tuple(x.split(" ")) for x in df.columns]
    parsed_variables = parsed_columns + [("time", "[UTC]")]
    df.columns = [x[0] for x in parsed_columns]
    ds = df.to_xarray()
    for var, unit in parsed_variables:
        ds[var].attrs["units"] = unit[1:-1]
    return ds


def whiffle_wind_metmast_csv_to_dataframe(csv_path: str) -> pd.DataFrame:
    """Load a whiffle wind csv with metmast data into a pandas dataframe.

    Parameters
    ----------
    csv_path : str
        Path to the whiffle wind csv file with simulated metmast data.

    Returns
    -------
    pd.DataFrame
        Dataframe with simulated metmast data.
    """
    df = _read_whiffle_wind_csv(csv_path)
    multi_height_cols, single_height_cols = _split_metmast_columns(df.columns)
    df_multi_height_vars = df[multi_height_cols]
    df_single_height_vars = df[single_height_cols]
    midx_multi_height_vars = df_multi_height_vars.columns.map(
        _parse_metmast_multi_height_column_name
    )
    midx_single_height_vars = df_single_height_vars.columns.map(
        _parse_metmast_single_height_column_name
    )
    midx = [
        (x[0], x[1], f"{x[2]} {x[3]}")
        for x in [*midx_multi_height_vars, *midx_single_height_vars]
    ]
    midx = pd.MultiIndex.from_tuples(midx, names=("metmast", "height [m]", None))
    df.columns = midx
    df = df.stack(level=[0, 1], future_stack=True).sort_index()
    df = df.rename(columns=WHIFFLE_WIND_BACKWARD_COMPATIBILITY_MAP)
    return df


def whiffle_wind_metmast_csv_to_dataset(csv_path: str) -> xr.Dataset:
    """Load a whiffle wind csv with metmast data into a xarray dataset.

    Parameters
    ----------
    csv_path : str
        Path to the whiffle wind csv file with simulated metmast data.

    Returns
    -------
    xr.Dataset
        Dataset with simulated metmast data.
    """
    df = whiffle_wind_metmast_csv_to_dataframe(csv_path)
    ds = metmast_dataframe_to_dataset(df)
    ds["time"] = ds.time.astype("datetime64[ns]")
    return ds


def metmast_dataframe_to_dataset(df: pd.DataFrame) -> xr.Dataset:
    """Convert a dataframe with metmast data into an xarray dataset.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe with simulated metmast data.

    Returns
    -------
    xr.Dataset
        Dataset with simulated metmast data.
    """
    df = df.rename_axis(("time", "metmast", "height"))
    parsed_columns = [tuple(x.split(" ")) for x in df.columns]
    parsed_variables = parsed_columns + [("time", "[UTC]"), ("height", "[m]")]
    df.columns = [x[0] for x in parsed_columns]
    multi_height_cols, single_height_cols = _split_metmast_columns(df.columns)
    ds = df[multi_height_cols].dropna(how="all").to_xarray()
    for single_height_df_var in single_height_cols:
        single_height_ds_var = single_height_df_var.split(" ")[0]
        ds[single_height_ds_var] = (
            df[single_height_ds_var].dropna().droplevel("height").to_xarray()
        )

    for var, unit in parsed_variables:
        ds[var].attrs["units"] = unit[1:-1]
    return ds


def windographer_csv_to_dataframe(
    csv_path: str,
    metmast_name: str = None,
    utc_time_diff: int = 0,
) -> pd.DataFrame:
    """Load a windographer csv containing metmast data into a pandas dataframe.

    Parameters
    ----------
    csv_path : str
        Path to the windographer csv file with measured metmast data.
    metmast_name : str, optional
        Name of the metmast.
    utc_time_diff : int, optional
        Difference in hours between UTC and local time.

    Returns
    -------
    pd.DataFrame
        Dataframe with measured metmast data.
    """
    df = pd.read_csv(csv_path, header=8, sep="\t", encoding="ISO-8859-1")
    df.index = pd.to_datetime(df["Date/Time"], dayfirst=True)
    df = df.drop("Date/Time", axis=1)
    df = pd.concat(
        [
            df.filter(like="AvgWd"),
            df.filter(like="AvgWs").filter(like="m/s"),
        ],
        axis=1,
    )
    df.columns = df.columns.map(_parse_windographer_column_name_to_tuple)
    df = df.sort_index(axis=1).replace(-999, np.nan)
    df = pd.concat(
        [
            df[["AvgWd"]]
            .T.groupby(level=[0, 1], group_keys=False)
            .apply(lambda x: mean_angle(x) if x.shape[0] > 1 else x.droplevel(2))
            .T,
            df[["AvgWs"]]
            .T.groupby(level=[0, 1], group_keys=False)
            .apply(lambda x: x.mean())
            .T,
        ],
        axis=1,
    )
    df = df.rename(columns=WINDOGRAPHER_COLUMN_MAP)
    df.index = (df.index - pd.Timedelta(utc_time_diff, unit="hour")).tz_localize("utc")
    df = df.stack(future_stack=True)
    df = pd.concat({metmast_name: df}, names=["metmast", "time [UTC]", "height [m]"])
    df = df.reorder_levels([1, 0, 2]).sort_index()
    return df


def cabauw_metmast_nc_to_dataset(netcdf_path: str):
    ds = xr.load_dataset(netcdf_path)
    ds = ds[list(CABAUW_DATA_VARS_MAP)]
    for var, (new_var, unit) in CABAUW_DATA_VARS_MAP.items():
        ds = ds.rename_vars({var: new_var})
        ds[new_var].attrs["units"] = unit
    for var, (new_var, unit) in CABAUW_DIMS_MAP.items():
        ds = ds.rename_dims({var: new_var}).rename_vars({var: new_var})
        ds[new_var].attrs["units"] = unit
    ds["time"] = ds["time"].dt.round("10min").to_index().tz_localize("utc")
    ds["time"].attrs["units"] = "UTC"
    ds = ds.assign_coords(metmast="cabauw_metmast").expand_dims("metmast")
    ds = ds.sel(height=CABAUW_HEIGHTS_TO_KEEP)
    return ds



from whiffle_client.forecast.mapping.asset_model import AssetEndpoints
from whiffle_client.forecast.mapping.forecast_model import ForecastModelEndpoints
from whiffle_client.forecast.mapping.observation_model import ObservationModelEndpoints



from pathlib import Path
from typing import Union

import yaml
import yaml_include

from whiffle_client.loaders.csv import csv_loader_constructor

CHUNK_SIZE = 1024 * 1024


def stream_download_write_chunks_url(filename, url, session):
    """Download files in streaming fashion

    This method avoids memory issues when downloading large files

    Parameters
    ----------
    filename : str
        String of file to use to dump data
    url : str
        Url to download from
    session
        request session to use to perform download
    """
    with session.get_request(url, stream=True) as res:
        stream_download_write_chunks(filename=filename, res=res)


def stream_download_write_chunks(filename, res):
    """Download files in streaming fashion

    This method avoids memory issues when downloading large files

    Parameters
    ----------
    filename : str
        String of file to use to dump data
    res : requests.Response
        Requests response object with data to download
    """
    with open(filename, "wb") as file:
        file_size = float(res.headers.get("Content-Length", "nan"))
        downloaded = 0
        for chunk in res.iter_content(chunk_size=CHUNK_SIZE):
            if chunk:
                file.write(chunk)
                downloaded += len(chunk)
                print(
                    "Downloaded {:>5.1%}\r".format(downloaded / file_size),
                    end="",
                    flush=True,
                )


def download_write_chunks(filename, res):
    file = open(filename, "wb")
    file_size = float(res.headers["Content-Length"])
    downloaded = 0
    for chunk in res.iter_content(chunk_size=CHUNK_SIZE):
        if chunk:
            file.write(chunk)
            downloaded += len(chunk)
            print(
                "Downloaded {:>5.1%}\r".format(downloaded / file_size),
                end="",
                flush=True,
            )
    file.close()


def load_yaml_with_include(
    filename: Union[Path, str], relative_to_file: bool = False
) -> dict:
    """Load yaml file with yaml and csv includes.

    Parameters
    ----------
    filename : Path | str
        Yaml file to load.
    relative_to_file : bool, optional
        Whether the path should be relative to the yaml file to load.
        If set to False, the path will be relative to the current working
        directory.

    Returns
    -------
    dict
        Yaml file content as python dictionary.
    """
    base_dir = None
    if relative_to_file:
        base_dir = Path(filename).resolve().parents[0]
    yaml.add_constructor("!include", yaml_include.Constructor(base_dir=base_dir))
    yaml.add_constructor("!include-csv", csv_loader_constructor(base_dir=base_dir))
    with open(filename) as f:
        data = yaml.load(f, Loader=yaml.FullLoader)
    del yaml.FullLoader.yaml_constructors["!include"]
    del yaml.FullLoader.yaml_constructors["!include-csv"]
    return data



from whiffle_client.wind.components.dates import Dates
from whiffle_client.wind.components.domain import Domain
from whiffle_client.wind.components.geometries import Geometries
from whiffle_client.wind.components.metmast import Metmast
from whiffle_client.wind.components.turbine import Turbine
from whiffle_client.wind.components.turbine_model_specs import TurbineModelSpecs
from whiffle_client.wind.components.wind_simulation_task import WindSimulationTask
from whiffle_client.wind.components.windfarm import Windfarm

__all__ = [
    "Dates",
    "Domain",
    "Geometries",
    "Metmast",
    "Turbine",
    "TurbineModelSpecs",
    "WindSimulationTask",
    "Windfarm",
]



import inspect
import logging
import json
from dataclasses import asdict, dataclass


logger = logging.getLogger(__name__)


@dataclass
class BaseComponent:
    @classmethod
    def from_dict(cls, env, input=False):
        instance = cls(
            **{k: v for k, v in env.items() if k in inspect.signature(cls).parameters}
        )
        instance._readonly_attrs_names = []
        for k, v in env.items():
            if k not in inspect.signature(cls).parameters:
                setattr(instance, k, v)
                instance._readonly_attrs_names.append(k)
                if input and k != "type":
                    logger.warn(
                        f"Input parameter {k} not defined. Please review if there is a typo and check examples for valid input parameters."
                    )

        return instance

    @classmethod
    def from_list(cls, env):
        instances = [cls.from_dict(sub_env) for sub_env in env]
        return instances

    def _get_api_params(self):
        return asdict(self)

    def __str__(self) -> str:
        data = json.dumps(
            asdict(self)
            | {k: getattr(self, k) for k in getattr(self, "_readonly_attrs_names", {})},
            indent=4,
        )
        return f"{self.__class__}:\n{data}"

    def __repr__(self) -> str:
        return json.dumps(
            asdict(self)
            | {k: getattr(self, k) for k in getattr(self, "_readonly_attrs_names", {})}
        )



import warnings

import numpy as np
import pandas as pd
import seaborn as sns
import xarray as xr
from matplotlib import pyplot as plt

from whiffle_client.analysis.skill import skill, skill_str
from whiffle_client.analysis.utils import (
    mean_wind_speed_by_hour_of_the_day,
    weibull_params_from_samples,
    weibull_pdf_from_params,
    wind_rose_frequency,
    wind_speed_histogram,
)


def _concat_wind_data(
    measured_wind_speed_series,
    aspire_wind_speed_series,
    concurrent,
):
    wind_speed_data = [
        measured_wind_speed_series.rename("measured"),
        aspire_wind_speed_series.rename("whiffle-wind"),
    ]
    wind_speed_df = pd.concat(wind_speed_data, axis=1)
    if concurrent:
        wind_speed_df = wind_speed_df.dropna()
    return wind_speed_df


def _wind_rose_set_yticks(ax, yticks=None):
    if yticks is not None:
        ax.set_yticks(yticks)

    labels = ["{:2.0f}%".format(item) for item in ax.get_yticks()]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ax.set_yticklabels(labels, fontsize=8)
    return ax


def _rose_dir_bins(nsector):
    angle = 360.0 / nsector
    dir_bin_centers = np.arange(-angle / 2, 360.0 + angle, angle, dtype=float)
    dir_bin_edges = list(dir_bin_centers)
    dir_bin_edges.pop(-1)
    dir_bin_edges[0] = dir_bin_edges.pop(-1)
    dir_bin_centers[0] = 0.0

    return dir_bin_centers, dir_bin_edges


def _rose_histogram(
    direction, var, nsector, wind_speed_bin_edges, normed=False, blowto=False
):
    """
    Returns an array where, for each sector of wind
    (centred on the north), we have the number of time the wind comes with a
    particular var (speed, polluant concentration, ...).
    * direction : 1D array - directions the wind blows from, North centred
    * var : 1D array - values of the variable to compute. Typically the wind
        speeds
    * wind_speed_bin_edges : list - list of var category against we're going to compute the table
    * nsector : integer - number of sectors
    * normed : boolean - The resulting table is normed in percent or not.
    * blowto : boolean - Normaly a windrose is computed with directions
    as wind blows from. If true, the table will be reversed (usefull for
    pollutantrose)

    """
    if len(var) != len(direction):
        raise (ValueError("var and direction must have same length"))

    dir_bin_centers, dir_bin_edges = _rose_dir_bins(nsector)

    var_bins = list(wind_speed_bin_edges)
    var_bins = [-np.inf] + var_bins + [np.inf]

    if blowto:
        direction = direction + 180.0
        direction[direction >= 360.0] = direction[direction >= 360.0] - 360

    table = np.histogram2d(
        x=var, y=direction, bins=[var_bins, dir_bin_centers], density=False
    )[0]
    # add the last value to the first to have the table of North winds
    table[:, 0] = table[:, 0] + table[:, -1]
    # and remove the last col
    table = table[:, :-1]
    if normed:
        table = table * 100 / table.sum()

    return dir_bin_edges, var_bins, table


def _plot_scatter_density_hist(
    x,
    y,
    ax=None,
    add_skill_str=True,
    fancy_skill_str=False,
    cmap=plt.cm.inferno,
    one_to_one=True,
    normalize=1.0,
):
    def _hist_colouring(
        xarray, yarray, nbins=1000, cmap=plt.cm.inferno, smoother=15, fnc=None
    ):
        """Helper function to generate a colored histogram."""

        def bin_indices(data, bins):
            ixs = np.zeros(len(data), int) - 1
            for i, bini in enumerate(bins[:-1]):
                inbin = (bini <= data) * (data < bins[i + 1])
                if i == len(bins) - 1:
                    inbin *= data == bins[i + 1]
                ixs += inbin * (i + 1)
            return ixs

        xarray = xarray[np.invert(np.isnan(xarray) | np.isnan(yarray))]
        yarray = yarray[np.invert(np.isnan(xarray) | np.isnan(yarray))]

        xbins = np.linspace(
            np.floor(np.min(xarray)) - 0.1, np.ceil(np.max(xarray)) + 0.1, nbins + 1
        )
        ybins = np.linspace(
            np.floor(np.min(yarray)) - 0.1, np.ceil(np.max(yarray)) + 0.1, nbins + 1
        )
        hist2d, _, _ = np.histogram2d(xarray, yarray, bins=[xbins, ybins])
        if fnc is not None:
            hist2d = fnc(hist2d)
        x_ix = bin_indices(xarray, xbins)
        y_ix = bin_indices(yarray, ybins)

        if smoother:
            smoothhist = np.zeros(len(x_ix))
            for i, (ix, iy) in enumerate(zip(x_ix, y_ix)):
                # smoothhist[i] = np.sum(hist2d[ix-smoother:ix+smoother+1,iy-smoother:iy+smoother+1])
                smoothhist[i] = np.mean(
                    hist2d[
                        max(0, ix - smoother) : ix + smoother + 1,
                        max(0, iy - smoother) : iy + smoother + 1,
                    ]
                )
            return cmap(smoothhist / np.max(smoothhist))
        else:
            return cmap(hist2d[x_ix, y_ix] / np.max(hist2d))

    inds = np.isfinite(x) & np.isfinite(y)

    x = x[inds]
    y = y[inds]

    cvals = _hist_colouring(x, y, nbins=1000, cmap=cmap, smoother=15, fnc=None)

    ax.scatter(x, y, c=cvals, edgecolor=None, s=1)

    my_skill_names = {"bias": "bias", "corr2": r"$R^2$", "MAE": "MAE", "RMSE": "RMSE"}

    def my_skill_str(a, b, skill_names, norm=None):
        msg = ""

        msg += "# of data points = {:d}\n".format(a.size)

        msg += "mean(a) = {:.4f}\n".format(float(np.nanmean(a)))
        msg += "mean(b) = {:.4f}\n".format(float(np.nanmean(b)))

        for score, value in skill(a, b, norm).items():
            if score in skill_names:
                msg += "{} = {:.4f}\n".format(skill_names[score], float(value))

        return msg.strip()

    if add_skill_str:
        if fancy_skill_str:
            skill_str_value = my_skill_str(
                x, y, my_skill_names, normalize
            )  # y is model
            props = dict(boxstyle="round", facecolor="white")

            ax.text(
                0.05,
                0.95,
                skill_str_value,
                transform=ax.transAxes,
                fontsize=10,
                verticalalignment="top",
                bbox=props,
            )

        else:
            skill_str_value = skill_str(y, x, norm=normalize)  # y is model

            props = dict(boxstyle="round", facecolor="white")

            ax.text(
                0.05,
                0.95,
                skill_str_value,
                transform=ax.transAxes,
                fontsize=6,
                verticalalignment="top",
                bbox=props,
            )

    if one_to_one:
        start = min(np.nanmin(x), np.nanmin(y))
        end = max(np.nanmax(x), np.nanmax(y))

        ax.plot([start, end], [start, end], "-k", zorder=4)

        ax.set_xlim(start, end)
        ax.set_ylim(start, end)

        ax.set_aspect(1)

    ax.set_xlabel("Measured wind speed [m/s]")
    ax.set_ylabel("Whiffle-wind wind speed [m/s]")
    ax.set_title("Wind speed scatter density plot with metrics")
    ax.grid()

    return ax


def plot_weibull_pdf_comparison(
    measured_wind_speed_series: pd.Series,
    aspire_wind_speed_series: pd.Series,
    concurrent: bool = True,
    ax: plt.axes = None,
) -> plt.axes:
    """Fit a Weibull distribution to the data and plot the PDF.

    Parameters
    ----------
    measured_wind_speed_series : pd.Series
        Measured wind speed series.
    aspire_wind_speed_series : pd.Series
        Simulated wind speed series.
    concurrent : bool, optional
        If True, compare concurrent wind speeds i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    ax : plt.axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()
    wind_speed_df = _concat_wind_data(
        measured_wind_speed_series,
        aspire_wind_speed_series,
        concurrent,
    )
    weibull_params_df = wind_speed_df.apply(
        lambda x: pd.Series(
            weibull_params_from_samples(x),
            index=["k", "A"],
        )
    )
    x_pdf = np.linspace(0, np.ceil(wind_speed_df.max().max()), 200)
    weibull_pdf_df = weibull_params_df.apply(
        lambda x: pd.Series(
            weibull_pdf_from_params(x_pdf, *x[["k", "A"]].tolist()),
            index=x_pdf,
        )
    )
    weibull_pdf_df.columns = [
        f'{c}:\nk={weibull_params_df[c]["k"]:.2f}\nA={weibull_params_df[c]["A"]:.2f} m/s'
        for c in weibull_pdf_df.columns
    ]
    weibull_pdf_df.plot(ax=ax)
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    ax.set_xlabel("Wind speed [m/s]")
    ax.set_ylabel("Weibull pdf [-]")
    ax.set_title("Fitted Weibull PDF comparison on wind speed")
    ax.grid()
    return ax


def plot_wind_speed_distribution_comparison(
    measured_wind_speed_series: pd.Series,
    aspire_wind_speed_series: pd.Series,
    concurrent: bool = True,
    bin_width: float = 1.0,
    kind: str = "bar",
    ax: plt.Axes = None,
) -> plt.Axes:
    """Compare wind speed distributions between measured and simulated data.

    Notes
    -----
    When 'kind' parameter is set to 'line', each point correspond to the
    histogram bin frequency (not the density). The 'line' plot represents
    the same underlying data as the 'bar' plot but is helpful for comparison.

    Parameters
    ----------
    measured_wind_speed_series : pd.Series
        Measured wind speed series.
    aspire_wind_speed_series : pd.Series
        Simulated wind speed series.
    concurrent : bool, optional
        If True, compare concurrent wind speeds i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    bin_width : float, optional
        Binning width.
    kind : str, optional
        Whether to plot distributions as 'bar' or 'line'.
    ax : plt.Axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    allowed_kind = ["bar", "line"]
    if kind not in allowed_kind:
        raise ValueError(f"'kind' parameter should be in {allowed_kind}")
    if ax is None:
        _, ax = plt.subplots()
    wind_speed_df = _concat_wind_data(
        measured_wind_speed_series,
        aspire_wind_speed_series,
        concurrent,
    )
    ws_hist_df = wind_speed_df.apply(wind_speed_histogram, bin_width=bin_width)
    ws_hist_df = ws_hist_df.fillna(0)
    ws_hist_df.index = ws_hist_df.index.map(lambda x: x.mid)
    ws_hist_df = ws_hist_df / ws_hist_df.sum()
    if kind == "line":
        ws_hist_df.plot(marker=".", ax=ax, legend=True)
    elif kind == "bar":
        ws_hist_df.plot.bar(ax=ax, legend=True)
    ax.set_xlabel("Wind speed bin center [m/s]")
    ax.set_ylabel("Frequency [-]")
    ax.set_title("Wind speed histogram comparison")
    ax.grid()
    return ax


def plot_wind_rose_frequency_comparison(
    measured_wind_direction_series: pd.Series,
    aspire_wind_direction_series: pd.Series,
    concurrent: bool = True,
    n_sectors: int = 12,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Compare sector frequencies between measured and simulated data.

    Notes
    -----
    This plot is in polar projection.

    Parameters
    ----------
    measured_wind_direction_series : pd.Series
        Measured wind direction series.
    aspire_wind_direction_series : pd.Series
        Simulated wind direction series.
    concurrent : bool, optional
        If True, compare concurrent wind directions i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    n_sectors : int, optional
        Number of sectors to use for the wind rose.
    ax : plt.Axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots(subplot_kw={"projection": "polar"})
    wind_dir_df = _concat_wind_data(
        measured_wind_direction_series,
        aspire_wind_direction_series,
        concurrent,
    )
    wind_rose_df = wind_dir_df.apply(
        wind_rose_frequency,
        n_sectors=n_sectors,
    )
    wind_rose_df = wind_rose_df / wind_rose_df.sum()
    wind_rose_df_rad = wind_rose_df.copy()
    wind_rose_df_rad.index = wind_rose_df_rad.index.map(lambda x: np.deg2rad(x.mid))
    for c in wind_rose_df_rad.columns:
        x = wind_rose_df_rad.index.to_series()
        x = pd.concat([x, x.iloc[[0]]]).to_numpy()
        y = wind_rose_df_rad[c]
        y = pd.concat([y, y.iloc[[0]]]).to_numpy()
        ax.plot(
            x,
            y,
            marker=".",
            label=c,
        )
    ax.set_theta_zero_location("N")
    ax.set_theta_direction(-1)
    ax.grid(True)
    ax.legend(loc="center left", bbox_to_anchor=(1.1, 0.5))
    ax.set_xlabel("Wind direction bin center [m/s]")
    ax.set_ylabel("Frequency [-]", labelpad=30)
    ax.set_title("Wind rose frequency comparison")
    return ax


def plot_wind_speed_scatter(
    measured_wind_speed_series: pd.Series,
    aspire_wind_speed_series: pd.Series,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Scatter plot of measured and simulated wind speeds.

    Parameters
    ----------
    measured_wind_speed_series : pd.Series
        Measured wind speed series.
    aspire_wind_speed_series : pd.Series
        Simulated wind speed series.
    ax : plt.Axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()
    wind_speed_df = _concat_wind_data(
        measured_wind_speed_series,
        aspire_wind_speed_series,
        concurrent=True,
    )
    ws_min = np.floor(wind_speed_df.min().min())
    ws_max = np.ceil(wind_speed_df.max().max())

    ax.set_aspect("equal")
    wind_speed_df.plot.scatter(
        x="measured",
        y="whiffle-wind",
        s=10,
        alpha=0.2,
        ax=ax,
    )
    ax.plot([ws_min, ws_max], [ws_min, ws_max], "-", color="k", lw=0.5)
    ax.set_xlim(ws_min, ws_max)
    ax.set_ylim(ws_min, ws_max)
    ax.set_xlabel("Measured wind speed [m/s]")
    ax.set_ylabel("Whiffle-wind wind speed [m/s]")
    ax.set_title("Wind speed scatter plot")
    ax.grid()
    return ax


def plot_wind_speed_scatter_density(
    measured_wind_speed_series: pd.Series,
    aspire_wind_speed_series: pd.Series,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Density plot of measured and simulated wind speeds.

    Notes
    -----
    This plot is particularly useful when the data contains many samples.

    Parameters
    ----------
    measured_wind_speed_series : pd.Series
        Measured wind speed series.
    aspire_wind_speed_series : pd.Series
        Simulated wind speed series.
    ax : plt.Axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()
    return _plot_scatter_density_hist(
        measured_wind_speed_series,
        aspire_wind_speed_series,
        ax=ax,
    )


def plot_wind_speed_diurnal_cycle_comparison(
    measured_wind_speed_series: pd.Series,
    aspire_wind_speed_series: pd.Series,
    concurrent: bool = True,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Compare mean diurnal cycle of measured and simulated wind speeds.

    Notes
    -----
    Each point in the plot represents the mean wind speed grouped
    by hour of the day.

    Parameters
    ----------
    measured_wind_speed_series : pd.Series
        Measured wind speed series.
    aspire_wind_speed_series : pd.Series
        Simulated wind speed series.
    concurrent : bool, optional
        If True, compare concurrent wind speeds i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    ax : plt.Axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()
    wind_speed_df = _concat_wind_data(
        measured_wind_speed_series,
        aspire_wind_speed_series,
        concurrent,
    )
    wind_speed_diurnal_cycle = wind_speed_df.apply(
        mean_wind_speed_by_hour_of_the_day,
    )
    wind_speed_diurnal_cycle.plot(marker=".", ax=ax)
    ax.set_xlabel("Hour of the day [-]")
    ax.set_ylabel("Mean wind speed by hour of the day [m/s]")
    ax.set_title("Wind speed diurnal cycle comparison")
    ax.grid()
    return ax


def plot_wind_speed_vertical_profile_comparison(
    measured_wind_speed_da: xr.DataArray,
    aspire_wind_speed_da: xr.DataArray,
    concurrent: bool = True,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Compare vertical profiles of measured and simulated wind speeds.

    Notes
    -----
        - Data arrays should contain time and height dimensions.
        - Data arrays can have different heights.
        - If 'concurrent' is True, time intersection between all heights
          and data arrays will be performed.

    Parameters
    ----------
    measured_wind_speed_da : xr.DataArray
        Measured wind speed data array.
    aspire_wind_speed_da : xr.DataArray
        Simulated wind speed data array.
    concurrent : bool, optional
        If True, compare concurrent wind speeds i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    ax : bool, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()
    measured_wind_speed_da = measured_wind_speed_da.dropna("height", how="all")
    if concurrent:
        measured_time_idx = measured_wind_speed_da.dropna("time")["time"].to_index()
        aspire_time_idx = aspire_wind_speed_da.dropna("time")["time"].to_index()
        concurrent_time_idx = measured_time_idx.intersection(aspire_time_idx)
        measured_wind_speed_da = measured_wind_speed_da.sel(time=concurrent_time_idx)
        aspire_wind_speed_da = aspire_wind_speed_da.sel(time=concurrent_time_idx)
    measured_wind_speed_da.mean("time", keep_attrs=True).plot(
        y="height",
        label="measured",
        marker=".",
        ax=ax,
    )
    aspire_wind_speed_da.mean("time", keep_attrs=True).plot(
        y="height",
        label="whiffle-wind",
        marker=".",
        ax=ax,
    )
    ax.legend()
    ax.set_xlabel("Mean wind speed [m/s]")
    ax.set_title("Mean wind speed vertical profile comparison")
    ax.grid()
    return ax


def plot_wind_speed_quantiles_vertical_profile_comparison(
    measured_wind_speed_da: xr.DataArray,
    aspire_wind_speed_da: xr.DataArray,
    concurrent: bool = True,
    quantiles: list[float] = [0.1, 0.5, 0.9],
    ax: plt.Axes = None,
) -> plt.Axes:
    """Compare vertical profiles of measured and simulated wind speeds.

    Notes
    -----
        - Data arrays should contain time and height dimensions.
        - Data arrays can have different heights.
        - If 'concurrent' is True, time intersection between all heights
          and data arrays will be performed.

    Parameters
    ----------
    measured_wind_speed_da : xr.DataArray
        Measured wind speed data array.
    aspire_wind_speed_da : xr.DataArray
        Simulated wind speed data array.
    concurrent : bool, optional
        If True, compare concurrent wind speeds i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    quantiles: list[float], optional
        Quantiles to plot.
    ax : bool, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()
    measured_wind_speed_da = measured_wind_speed_da.dropna("height", how="all")
    if concurrent:
        measured_time_idx = measured_wind_speed_da.dropna("time")["time"].to_index()
        aspire_time_idx = aspire_wind_speed_da.dropna("time")["time"].to_index()
        concurrent_time_idx = measured_time_idx.intersection(aspire_time_idx)
        measured_wind_speed_da = measured_wind_speed_da.sel(time=concurrent_time_idx)
        aspire_wind_speed_da = aspire_wind_speed_da.sel(time=concurrent_time_idx)
    data_da = xr.concat(
        [
            measured_wind_speed_da.assign_coords(source="measured"),
            aspire_wind_speed_da.assign_coords(source="whiffle-wind"),
        ],
        dim="source",
    )
    quantiles_df = data_da.quantile(quantiles, dim="time").to_series().reset_index()
    ax = sns.lineplot(
        data=quantiles_df,
        x="horizontal-wind-speed",
        y="height",
        hue="source",
        style="quantile",
        marker=".",
        markeredgecolor=None,
        orient="y",
        ax=ax,
    )
    ax.legend()
    ax.set_xlabel("Wind speed quantiles [m/s]")
    ax.set_ylabel("Height [m]")
    ax.set_title("Wind speed quantiles vertical profile comparison")
    ax.grid()
    return ax


def plot_wind_rose(
    wind_direction_series: pd.Series,
    wind_speed_series: pd.Series,
    wind_speed_bin_edges: np.array = np.arange(0, 30, 5),
    nsector: int = 16,
    varunit: str = "m/s",
    cmap: str = "viridis",
    ax: plt.Axes = None,
) -> plt.Axes:
    """Plot a colored wind rose.

    Parameters
    ----------
    wind_direction_series : pd.Series
        Wind direction.
    wind_speed_series : pd.Series
        Wind speed.
    wind_speed_bin_edges : np.array, optional
        Bin edges for color bins of wind_speed_series.
    nsector : int, optional
        Number of sector for the wind rose.
    varunit : str, optional
        Unit for wind speed.
    cmap : str, optional
        Matplotlib colormap to use.
    ax : plt.Axes, optional
        Axis for the plot.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots(subplot_kw={"projection": "polar"})

    dir_bin_edges, var_bins, table = _rose_histogram(
        wind_direction_series,
        wind_speed_series,
        nsector,
        wind_speed_bin_edges,
        normed=True,
    )

    nbins = len(var_bins) - 1
    colors = plt.get_cmap(cmap)(np.linspace(0.1, 0.9, nbins))

    width = 0.8 * 2 * np.pi / nsector
    angles = np.arange(0, -2 * np.pi, -2 * np.pi / nsector) + np.pi / 2

    offset = np.zeros(nsector)
    for bin in range(nbins):
        zorder = -1000 + nbins - bin
        if bin < 1:
            binlabel = "< {:2.1f} {}".format(var_bins[1], varunit)
        elif bin < nbins - 1:
            binlabel = "{:2.1f}-{:2.1f} {}".format(
                var_bins[bin], var_bins[bin + 1], varunit
            )
        else:
            binlabel = "> {:2.1f} {}".format(
                var_bins[bin], varunit
            )  # adjust final bin label
        ax.bar(
            angles,
            table[bin, :],
            width=width,
            color=colors[bin],
            bottom=offset,
            zorder=zorder,
            linewidth=0,
            label=binlabel,
        )
        offset += table[bin, :]

    theta_angles = np.arange(0, 360, 45)
    theta_labels = ["E", "N-E", "N", "N-W", "W", "S-W", "S", "S-E"]
    ax.set_thetagrids(
        angles=theta_angles, labels=theta_labels
    )  # Fails if you don't pass polar-projected ax

    ax = _wind_rose_set_yticks(ax)

    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    ax.legend(loc="upper left", bbox_to_anchor=(1.05, 1.0))

    return ax


def plot_timeseries_comparison(
    measured_series: pd.Series,
    aspire_series: pd.Series,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Plot time series comparison.

    Parameters
    ----------
    measured_series : pd.Series
        Measured time series.
    aspire_series : pd.Series
        Whiffle wind time series.
    ax : plt.Axes, optional
        Axis for the plot.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()

    if measured_series.name != aspire_series.name:
        raise ValueError("Series names must match.")

    ax = measured_series.plot(ax=ax, label="measured")
    ax = aspire_series.plot(ax=ax, label="whiffle-wind")
    ax.legend()
    ax.set_xlabel("time")
    ax.set_ylabel(measured_series.name)
    ax.set_title("Wind speed comparison")
    ax.grid()
    return ax



import numpy as np


def cp_to_power(wind_speed, power_coefficient, radius, reference_density=1.225):
    return (
        0.5
        * power_coefficient
        * np.pi
        * radius**2.0
        * reference_density
        * wind_speed**3.0
    )


def ct_to_thrust(wind_speed, thrust, radius, reference_density=1.225):
    return 0.5 * thrust * np.pi * radius**2.0 * reference_density * wind_speed**2.0


def create_mock_design_curves(
    name,
    hub_height,
    rotor_diameter,  # [mm]
    rated_power,  # [W]
    rated_wind_speed=11,  # [m/s]
    cutin_wind_speed=4,  # [m/s]
    cutout_wind_speed=24,  # [m/s]
    reference_density=1.225,  # [kg/m^3]
    reference_turbulence_intensity=0.08,  # [-]
):
    reference_windspeed = np.arange(cutin_wind_speed, cutout_wind_speed + 0.5, 0.5)

    # Compute power
    power = np.minimum(
        cp_to_power(reference_windspeed, 0.45, rotor_diameter / 2, reference_density),
        rated_power,
    )
    # NOTE: it might reach rated power before rated_wind_speed ("v164_9500" example)
    # https://www.thewindpower.net/turbine_en_1476_vestas_v164-9500.php
    power[reference_windspeed >= rated_wind_speed] = rated_power

    # Compute thrust_coefficient
    thrust = ct_to_thrust(
        reference_windspeed, 0.9, rotor_diameter / 2, reference_density
    )
    before_rated_wind_speed = rated_wind_speed - 1
    max_thrust = thrust[reference_windspeed == before_rated_wind_speed]
    windspeed_mask = reference_windspeed > before_rated_wind_speed
    thrust[windspeed_mask] = max_thrust * np.maximum(
        0.1,
        (
            1
            - (reference_windspeed[windspeed_mask] - before_rated_wind_speed)
            / (cutout_wind_speed - before_rated_wind_speed)
        ),
    )

    thrust_coefficient = thrust / ct_to_thrust(
        reference_windspeed, 1, rotor_diameter / 2, reference_density
    )

    return {
        "name": name,
        "hub_height": hub_height,
        "rotor_diameter": rotor_diameter,
        "rated_power": rated_power,  # [W]
        "reference_density": reference_density,
        "reference_turbulence_intensity": reference_turbulence_intensity,
        "reference_windspeed": reference_windspeed.round(4).tolist(),
        "thrust_coefficient": thrust_coefficient.round(4).tolist(),
        "power": power.round(4).tolist(),
        "public": False,  # Expose turbine to other users
    }



import datetime
from dataclasses import dataclass, field

from whiffle_client.common.components.base import BaseComponent


# pylint: disable=duplicate-code
@dataclass
class ObservationModel(BaseComponent):
    """Observation model"""

    time: datetime.datetime = field(default=None)
    quantity: str = field(default=None)
    value: float = field(default=None)
    asset_name: int = field(default=None)



import numpy as np
import pandas as pd
from scipy.stats import exponweib


def wind_speed_histogram(
    wind_speed_series: pd.Series,
    bin_width: float = 1.0,
) -> pd.Series:
    """Calculate wind speed histogram from wind speed series.

    Parameters
    ----------
    wind_speed_series : pd.Series
        Wind speed series.
    bin_width : float, optional
        Bin width

    Returns
    -------
    pd.Series
        Wind speed histogram with IntervalIndex.
    """
    ws_min = 0
    ws_max = np.ceil(wind_speed_series.max())
    bins = np.arange(ws_min, ws_max + 1, bin_width) - bin_width / 2
    ws_cut = pd.cut(wind_speed_series, bins=bins)
    ws_hist = ws_cut.value_counts().sort_index()
    return ws_hist


def wind_rose_frequency(
    wind_direction_series: pd.Series,
    n_sectors: int = 12,
) -> pd.Series:
    """Calculate wind rose frequency from wind direction series.

    Parameters
    ----------
    wind_direction_series : pd.Series
        Wind direction series.
    n_sectors : int, optional
        Number of sectors for the wind rose.

    Returns
    -------
    pd.Series
        Wind rose frequency with IntervalIndex.
    """
    dir_bin_width = 360 / n_sectors
    dir_bins = np.linspace(0, 360 + dir_bin_width, n_sectors + 2)
    dir_bins = dir_bins - dir_bin_width / 2
    dir_intervals = pd.IntervalIndex.from_breaks(dir_bins)
    wd_cut = pd.cut(wind_direction_series, bins=dir_intervals)
    wd_hist = wd_cut.value_counts().sort_index()
    wd_hist[dir_intervals[0]] = wd_hist[dir_intervals[0]] + wd_hist[dir_intervals[-1]]
    wd_hist.drop(dir_intervals[-1], inplace=True)
    return wd_hist


def mean_wind_speed_by_hour_of_the_day(
    wind_speed_series: pd.Series,
) -> pd.Series:
    """Mean wind speed by hour of the day.

    Parameters
    ----------
    wind_speed_series : pd.Series
        Wind speed series.

    Returns
    -------
    pd.Series
        Mean wind speed by hour of the day.
    """
    return wind_speed_series.groupby(wind_speed_series.index.hour).mean()


def weibull_params_from_samples(data: pd.Series) -> tuple[float, float]:
    """Estimate Weibull parameters from samples.

    Parameters
    ----------
    data : pd.Series
        Data samples.

    Returns
    -------
    tuple[float, float]
        Shape and scale parameters of the Weibull distribution.
    """
    _, k, _, A = exponweib.fit(data.dropna(), fa=1, floc=0)
    return k, A


def weibull_pdf_from_params(
    x_pdf: pd.Series,
    k: float,
    A: float,
) -> pd.Series:
    """Estimate PDF of a Weibull distribution.

    Parameters
    ----------
    x_pdf : pd.Series
        Random variable values.
    k : float
        Scale parameter.
    A : float
        Shape parameter.

    Returns
    -------
    pd.Series
        PDF of the Weibull distribution at `x_pdf`.
    """
    return exponweib.pdf(x_pdf, 1, k, 0, A)


def mean_angle(a, weight=None, axis=None):
    """
    When averaging wind directions, `weight` is the wind speed
    """
    c = np.pi / 180

    if weight is None:
        weight = np.ones(a.shape)

    xs = np.cos(a * c) * weight
    ys = np.sin(a * c) * weight

    return np.arctan2(np.nanmean(ys, axis=axis), np.nanmean(xs, axis=axis)) / c % 360



#!/usr/bin/env python3

from whiffle_client.base_client import BaseClient
from whiffle_client.forecast.mapping import (
    AssetEndpoints,
    ForecastModelEndpoints,
    ObservationModelEndpoints,
)


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-few-public-methods
class WhiffleForecastClient:
    """
    Client to connect to the Whiffle Forecast API
    """

    # NOTE: https://refactoring.guru/design-patterns/facade
    def __init__(self, access_token=None, url=None, session=None):
        self.session = BaseClient(access_token, url, session)

        # Map the asset simulation model endpoints
        self.asset_model = AssetEndpoints(self.session)
        self._map_asset_model_specs_methods()

        # Map the observation model specs endpoints
        self.observation_model = ObservationModelEndpoints(self.session)
        self._map_observation_model_specs_methods()

        # Map the forecast model specs endpoints
        self.forecast_model = ForecastModelEndpoints(self.session)
        self._map_forecast_model_specs_methods()

    def __repr__(self) -> str:
        return self.session.__repr__()

    def _map_asset_model_specs_methods(self):
        self.add_asset_model = self.asset_model.add
        self.get_all_asset_model = self.asset_model.get_all
        self.get_asset_model = self.asset_model.get

    def _map_observation_model_specs_methods(self):
        self.add_observation_model = self.observation_model.add
        self.get_all_observation_model = self.observation_model.get_all

    def _map_forecast_model_specs_methods(self):
        self.add_forecast_model = self.forecast_model.add
        self.get_forecast_model = self.forecast_model.get



from whiffle_client.forecast.components.asset_model import AssetModel
from whiffle_client.forecast.components.forecast_model import ForecastModel
from whiffle_client.forecast.components.observation_model import ObservationModel

__all__ = ["AssetModel", "ObservationModel", "ForecastModel"]



from dataclasses import dataclass, field

from whiffle_client.common.components.base import BaseComponent


@dataclass
class AssetModel(BaseComponent):
    """Asset model"""

    # Metadata attributes
    name: str = field(default=None)
    location: dict[str, list[float]] = field(default=None)
    properties: list = field(default_factory=list)



from pathlib import Path
from typing import Union

from whiffle_client.forecast.components.forecast_model import ForecastModel
from whiffle_client.decorators import load_data
from whiffle_client.common.mapping.base import BaseMapping


class ForecastModelEndpoints(BaseMapping):
    """Forecast model endpoint

    Parameters
    ----------
    BaseMapping : Basemapping
        Base mapping
    """

    URL = "/api/v1/forecasts"
    RESOURCE_TYPE = "ForecastModel"

    # Forecast model commands
    @load_data(ForecastModel)
    def add(
        self,
        data: Union[str, dict, Path, ForecastModel] = None,
    ) -> ForecastModel:
        """Add new Forecast Model

        Parameters
        ----------
        data : Union[str, dict, Path, ForecastModel], optional
            Either a path to a yaml, the data itself as a dictionary or a ForecastModel object
            containing the parameters that define the forecast model specs, by default None

        Returns
        -------
        ForecastModel
            Object instance of the forecast  model
        """
        request = self.session.post_request(
            f"{self.session.server_url}{self.URL}", data=data
        )
        return ForecastModel.from_dict(request.json())

    # pylint: disable=arguments-differ
    def get(
        self,
        asset_name: str,
        name: str = "",
        reference_time: str = "",
    ):
        """Get Forecasts

        Parameters
        ----------
        reference_time : str
            The timestamp of the forecast to get
        asset_name : str
            The name of the asset to get the forecast for
        name : str, optional
            Forecast name, by default "name"
        """
        params = {}
        for key, value in [
            ("asset_name", asset_name),
            ("name", name),
            ("reference_time", reference_time),
        ]:
            if value != "":
                params[key] = value

        url = f"{self.session.server_url}{self.URL}"
        request = self.session.get_request(
            url,
            params=params,
        )
        return ForecastModel.from_dict(request.json())

    def get_all(self):
        """Abstract method for get_all

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def edit(self):
        """Abstract method for edit

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def delete(self):
        """Abstract method for delete

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def download(self):
        """Abstract method for download

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError



from datetime import datetime

import click
import yaml

from whiffle_client import Client
from whiffle_client.forecast import WhiffleForecastClient


@click.group()
@click.version_option(prog_name="whiffle-client")
def whiffle():
    pass


@whiffle.group()
def wind():
    """Wind api commands"""
    ...


@whiffle.group()
def task():
    """Task commands"""
    pass


@whiffle.group()
def config():
    """Edit config"""
    pass


@task.command()
@click.argument("number_of_tasks", default=10)
def list(number_of_tasks):
    """List tasks"""

    number_of_tasks = int(number_of_tasks)
    client = Client()
    tasks = client.get_tasks()
    click.echo(
        "{:33}{:11}{:16} {:16} progress".format(
            "task id", "status", "start time", "finish time"
        )
    )
    for task in tasks[-number_of_tasks:]:
        finished = task["finished"]
        if finished == None:
            task["finished"] = ""
        click.echo(
            "{task_id:33}{task_status:11}{received:17.16}{finished:17.16}{processed_steps}/{total_steps}".format(
                **task
            )
        )


@task.command()
@click.argument("file_path")
def run(file_path) -> str:
    """Run task type given a set of parameters.

    \b
    Parameters
    ----------
    file_path : str
        Path to (json|yaml) file containing parameters

    \b
    Returns
    -------
    str
        Name of the launched task
    """
    click.echo(f"run task with params: file_path={file_path}")
    client = Client()
    return client.process(file_path)


@config.command()
def login() -> str:
    """Log in into Whiffle's portal"""
    # Retrieve token from Whiffle's portal
    client = Client()
    client.get_token()
    click.echo(f"Token properly updated")


@config.command()
@click.argument("key")
@click.argument("value")
def edit(key, value) -> str:
    """Edit Whiffle configuration file

    \b
    Parameters
    ----------
    key : str
        Dot separated key value (e.g.: user.token)
    value : str
        Value to fill the configuration key with
    """
    client = Client()
    config_key = key.split(".")
    config = client.get_config()

    sub_config = config
    for sub_key in config_key[:-1]:
        if sub_key in sub_config:
            sub_config = sub_config[sub_key]
        else:
            raise KeyError(f"Key {sub_key} does not exist")
    try:
        if config_key[-1] not in sub_config:
            raise KeyError
        sub_config[config_key[-1]] = value
    except KeyError:
        raise KeyError(f"Key {config_key[-1]} does not exist")

    client.set_config(config=config)
    click.echo(f"{key} changed to {value}")


@config.command()
def list():
    """List configuration values"""

    client = Client()
    config = client.get_config()

    click.echo("Current configuration:\n-----\n")
    click.echo(yaml.safe_dump(config))
    click.echo("-----")


@task.command()
@click.argument("task_id")
def download(task_id):
    """Download task

    \b
    Parameters
    ----------
    task_id : str
    """
    click.echo(f"downloading task output {task_id}")
    client = Client()
    client.download(task_id)


@task.command()
@click.argument("task_id")
def attach(task_id):
    """attach task"""
    click.echo(f"attaching to the task {task_id}")
    client = Client()
    client.communicate(task_id)


@task.command()
@click.argument("task_id")
def cancel(task_id):
    """cancel task"""
    click.echo(f"cancelling task {task_id}")
    client = Client()
    client.cancel(task_id)


# Whiffle wind API commands


# Forecast group
@whiffle.group()
def forecast():
    """Forecast commands"""
    # pylint: disable=unnecessary-pass
    pass


@forecast.command()
@click.argument("number_of_assets", default=10)
def list_assets(number_of_assets):
    """List assets

    \b
    Parameters
    ----------
    number of assets : int
        Number of assets you want to see, by default 10

    \b
    Returns
    -------
    dict
        Table format of listed assets"""

    number_of_assets = int(number_of_assets)
    client = WhiffleForecastClient()
    assets = client.get_all_asset_model()
    name = "name"
    location = "location"
    click.echo(f"{name:33}{location:11}")
    for asset in assets[-number_of_assets:]:
        name = asset.name
        location = asset.location["coordinates"]
        click.echo(f"{name:33}{location}")


@forecast.command()
@click.argument("file_path")
def add_asset(file_path) -> str:
    """Add asset given a set of parameters.

    \b
    Parameters
    ----------
    file_path : str
        Path to (json|yaml) file containing parameters

    \b
    Returns
    -------
    dict
        JSON of added assets
    """
    click.echo(f"Add asset with params: file_path={file_path}")
    client = WhiffleForecastClient()
    return client.add_asset_model(**{"data": file_path})


@forecast.command()
@click.argument("asset_name")
def get_asset(asset_name):
    """Get asset

    \b
    Parameters
    ----------
    asset_name : str
        The name of the asset

    \b
    Returns
    -------
    dict
        JSON of asset"""

    client = WhiffleForecastClient()
    asset = client.get_asset_model(asset_name)
    name = "name"
    location = "location"
    click.echo(f"{name:33}{location:11}")
    name = asset.name
    location = asset.location["coordinates"]
    click.echo(f"{name:33}{location}")


@forecast.command()
@click.argument("asset_name")
@click.argument("quantity")
@click.argument("time_after")
@click.argument("time_before")
def list_observations(asset_name, quantity, time_before, time_after):
    """List observations

    \b
    Parameters
    ----------
    asset_name : str
        The name of the asset
    quantity : str
        The quantity of the observation, choose from: "PRODUCTION" "PRODUCTION_RAW"
        "SETPOINT_PLANNED", "CURTAILMENT_PLANNED", "REDISPATCH_PLANNED", "AVAILABILITY_PLANNED",
        "SETPOINT_REALIZED", "CURTAILMENT_REALIZED", "REDISPATCH_REALIZED", "AVAILABILITY_REALIZED",
        "WIND_SPEED", "WIND_DIRECTION", "TEMPERATURE" and "TOTAL_INPLANE_IRRADIANCE"
    time_after : str
        The time after which observations need to be listed
    time_before : str
        The time before which observations need to be listed

    \b
    Returns
    -------
    dict
        Table format of the observations"""
    time_after = datetime.fromisoformat(time_after)
    time_before = datetime.fromisoformat(time_before)
    client = WhiffleForecastClient()
    observations = client.get_all_observation_model(
        asset_name=asset_name,
        quantity=quantity.upper(),
        time_before=time_before,
        time_after=time_after,
    )

    click.echo(f"{'time':24}{'asset_name':24}{'quantity':24}{'value':10}")
    for observation in observations:
        asset_name = observation.asset_name
        quantity = observation.quantity
        time = observation.time
        value = observation.value
        click.echo(f"{time:24}{asset_name:24}{quantity:24}{value:10}")


@forecast.command()
@click.argument("file_path")
def add_observation(file_path) -> str:
    """Add observation(s) given a set of parameters.

    \b
    Parameters
    ----------
    file_path : str
        Path to (json|yaml) file containing parameters

    \b
    Returns
    -------
    dict
        JSON of the added observations
    """
    click.echo(f"Add observation with params: file_path={file_path}")
    client = WhiffleForecastClient()
    return client.add_observation_model(**{"data": file_path})


@forecast.command()
@click.argument("asset_name")
@click.argument("name", default="")
@click.argument("reference_time", default="")
def get_forecast(name, asset_name, reference_time):
    """Get forecast

    \b
    Parameters
    ----------
    asset_name : str
        The name of the asset
    name : str
        The name of the forecast (by default: blender)
    reference_time : str (isoformat)
        The reference time of the forecast (by default a get will be done on the latest forecast)

    \b
    Returns
    -------
    dict
        Table format of the forecast"""
    if reference_time not in ("latest", ""):
        reference_time = datetime.fromisoformat(reference_time).isoformat()
    client = WhiffleForecastClient()
    forecast_model = client.get_forecast_model(
        name=name, asset_name=asset_name, reference_time=reference_time
    )

    click.echo(
        f"Forecast: {forecast_model.name} with reference_time {forecast_model.reference_time} for asset "
        f"{forecast_model.asset_name}, results:"
    )
    click.echo(f"{'time':24}{'quantity':24}{'value':10}")
    for forecaststep in forecast_model.steps:
        quantity = forecaststep["quantity"]
        time = forecaststep["time"]
        value = forecaststep["value"]
        click.echo(f"{time:24}{quantity:24}{value:10}")


@forecast.command()
@click.argument("file_path")
def add_forecast(file_path) -> str:
    """Add forcast given a set of parameters.

    \b
    Parameters
    ----------
    file_path : str
        Path to (json|yaml) file containing parameters

    \b
    Returns
    -------
    dict
        JSON of the added forecast
    """
    click.echo(f"Add forecast with params: file_path={file_path}")
    client = WhiffleForecastClient()
    return client.add_forecast_model(**{"data": file_path})



import os
import secrets
import time
import urllib
import webbrowser
from pathlib import Path
from time import sleep

import pkg_resources
import platformdirs
import requests
import yaml
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from whiffle_client.auth import (
    AUTH_STATE,
    ServerThread,
    auth0_url_encode,
    generate_challenge,
)
from whiffle_client.decorators import request_ok, with_token


class BaseClient:
    """
    Base class client to connect to Whiffle APIs
    """

    # API variables
    ENDPOINTS_URL: str = ""

    CONFIG_FILE_NAME = "whiffle_config.yaml"
    CONFIG_PACKAGE_FILE_LOCATION = pkg_resources.resource_filename(
        "whiffle_client", f"resources/{CONFIG_FILE_NAME}"
    )  # package resource path
    CONFIG_USER_FILE_LOCATION = (
        f"{platformdirs.user_config_dir('whiffle')}/{CONFIG_FILE_NAME}"
    )

    config: dict = {}

    # Type method
    def __init__(self, access_token=None, token_config=None, url=None, session=None):
        """
        Initialize the client.

        Authentication order:
        1. `access_token` passed when creating class.
        2. `token_config` passed when creating class.
        3. token in CONFIG_FILE_PATH_LOCATION (JSON format)

        Parameters
        ----------
        access_token : str, optional
            Token for client session auth
        token_config : dict(str), optional
            Token configuration for client session auth
        url : str, optional
            Url pointing to API
        """

        self.config = self.get_config()
        if access_token:
            self.update_token_config(
                token={"access_token": access_token}, persist=False
            )
        if token_config:
            self.update_token_config(token=token_config, persist=False)

        access_token_validator = self.config["user"].get("token_validator", None)
        access_token = self.config["user"]["access_token"]

        url = url or self.config["whiffle"]["url"]
        self.server_url = url

        if session is None:
            # More docs: https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.Retry
            status_forcelist = (500, 502, 503, 504)
            retry = Retry(
                total=5,  # Total number of retries to allow
                backoff_factor=0.1,  # Incremental time between retry requests
                status_forcelist=status_forcelist,  # A set of integer HTTP status codes that will force a retry on
            )
            adapter = HTTPAdapter(max_retries=retry)
            session = requests.Session()
            session.mount("http://", adapter)
            session.mount("https://", adapter)
        self.session = session

        self.session.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {access_token}",
            "Authorization-type": (
                "Integration" if access_token_validator == "zitadel" else None
            ),
        }

    def __repr__(self) -> str:
        return f"Whiffle wind client connected to url: {self.server_url}"

    def update_token_config(self, token, persist=True):
        self.config = self.get_config()
        self.config["user"]["access_token"] = token["access_token"]
        self.config["user"]["refresh_token"] = token.get("refresh_token", "")
        self.config["user"]["token_validator"] = token.get("token_validator", "manual")
        self.config["user"]["expires_in"] = token.get("expires_in", 3600)
        self.config["user"]["obtained_at"] = token.get("obtained_at", int(time.time()))
        if persist:
            self.set_config(config=self.config)

    def get_token(self):
        IDP_BASE_URL = self.config["whiffle"]["idp_base_url"]
        CLIENT_ID = self.config["whiffle"]["client_id"]
        REDIRECT_URI = self.config["whiffle"]["redirect_uri"]

        verifier = auth0_url_encode(secrets.token_bytes(32))
        challenge = generate_challenge(verifier)
        base_url = IDP_BASE_URL + "/authorize?"
        url_parameters = {
            "client_id": CLIENT_ID,
            "redirect_uri": REDIRECT_URI,
            "scope": "profile openid email urn:zitadel:iam:user:resourceowner offline_access",
            "response_type": "code",
            "response_mode": "query",
            "code_challenge_method": "S256",
            "code_challenge": challenge.replace("=", ""),
            "state": auth0_url_encode(secrets.token_bytes(32)),
        }
        # Open the browser window to the login url
        webbrowser.open_new(base_url + urllib.parse.urlencode(url_parameters))
        server = ServerThread()
        # Start the server
        server.start()
        print(f"Waiting for log in callback")
        print(
            f"If no browser opens, please visit: {base_url}{urllib.parse.urlencode(url_parameters)}"
        )

        # Poll until the callback has been invoked or timeout
        waited = 2
        sleep(2)  # Usually in two seconds token has been refreshed
        while not AUTH_STATE["received_callback"] and waited < 60:
            sleep(1)
            waited += 1
            print(
                f"Waiting for callback for {waited} seconds, status {AUTH_STATE['received_callback']}",
                end="\r",
                flush=True,
            )
        if not AUTH_STATE["received_callback"]:
            raise RuntimeError("Timeout waiting for log in callback")
        server.shutdown()

        # received_state and error_message are global variable from auth_app callback
        if url_parameters["state"] != AUTH_STATE["received_state"]:
            raise RuntimeError(
                "Error: session replay or similar attack in progress. Please log out of all connections."
            )
        if AUTH_STATE["error_message"]:
            raise RuntimeError(
                "An error occurred: {}".format(AUTH_STATE["error_message"])
            )
        # Exchange the code for a token
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        body = {
            "grant_type": "authorization_code",
            "code": AUTH_STATE["code"],  # global variable from auth_app callback
            "redirect_uri": REDIRECT_URI,
            "client_id": CLIENT_ID,
            "code_verifier": verifier,
        }
        token_request = requests.post(
            IDP_BASE_URL + "token", headers=headers, data=body
        )
        if token_request.status_code != 200:
            raise RuntimeError(
                f"Error: failed to exchange code for token {token_request.status_code}"
            )

        token_request = token_request.json()
        token_request["token_validator"] = "zitadel"
        self.update_token_config(token_request, persist=True)

    def refresh_token(self):
        IDP_BASE_URL = self.config["whiffle"]["idp_base_url"]
        CLIENT_ID = self.config["whiffle"]["client_id"]

        if (
            "refresh_token" not in self.config["user"]
            or not self.config["user"]["refresh_token"]
        ):
            raise RuntimeError("No refresh token available, must log in again.")

        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = {
            "grant_type": "refresh_token",
            "refresh_token": self.config["user"]["refresh_token"],
            "client_id": CLIENT_ID,
        }
        token_refresh_response = requests.post(
            IDP_BASE_URL + "token", headers=headers, data=data
        )

        if token_refresh_response.status_code != 200:
            raise RuntimeError(
                f"Failed to refresh token: {token_refresh_response.status_code} {token_refresh_response.text}"
            )

        token_refresh_response = token_refresh_response.json()
        token_refresh_response["token_validator"] = "zitadel"
        self.update_token_config(token_refresh_response, persist=True)

    def get_valid_access_token(self):
        """Return a valid access token, refreshing if needed."""
        # NOTE: use already provided config or load it again
        if not self.config["user"]["access_token"]:
            self.get_token()
            self.config = self.get_config()

        if (
            time.time()
            > self.config["user"]["obtained_at"]
            + self.config["user"]["expires_in"]
            - 30
        ):
            self.refresh_token()
            self.config = self.get_config()

        return self.config["user"]["access_token"]

    def merge_configurations(self, base_config: dict, update_config: dict) -> dict:
        for key, value in update_config.items():
            if key not in base_config:
                base_config[key] = value
            elif isinstance(value, dict) and isinstance(base_config.get(key), dict):
                self.merge_configurations(base_config[key], value)

        return base_config

    def get_config(self) -> dict:
        """Gathers client configuration from local and updates with resources one.

        Returns
        -------
        dict
            Dictionary containing the configuration.

        Raises
        ------
        FileNotFoundError
            Raises error if no configuration found
        """
        with open(self.CONFIG_PACKAGE_FILE_LOCATION) as file_object:
            config = yaml.safe_load(file_object)

        # Merge default config with user config if present
        user_config = {}
        if Path(self.CONFIG_USER_FILE_LOCATION).exists():
            with open(self.CONFIG_USER_FILE_LOCATION) as file_object:
                user_config = yaml.safe_load(file_object)

        return self.merge_configurations(user_config, config)

    def set_config(self, config):
        config_file_path = self.CONFIG_USER_FILE_LOCATION
        os.makedirs(os.path.dirname(config_file_path), exist_ok=True)
        with open(config_file_path, "w") as file_object:
            yaml.safe_dump(config, file_object)

    @request_ok
    @with_token
    def get_request(self, url, **kwargs):
        return self.session.get(url, **kwargs)

    @request_ok
    @with_token
    def post_request(self, url, data=None, files=None):
        if files:
            return self.session.post(url, data=data, files=files)

        return self.session.post(url, json=data)

    @request_ok
    @with_token
    def put_request(self, url, data=None, files=None):
        if files:
            return self.session.put(url, data=data, files=files)

        return self.session.put(url, json=data)

    @request_ok
    @with_token
    def delete_request(self, url):
        return self.session.delete(url)



from pathlib import Path

from functools import wraps

from whiffle_client.io import load_yaml_with_include


def load_data(class_type, resource_type=None):
    def wrap(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            data = kwargs.get("data", None)

            # Set data if lists are provided
            if data is None and len(args) == 1:
                data = args[0]
                args = ()

            if isinstance(data, str) or isinstance(data, Path):
                data = load_yaml_with_include(data, relative_to_file=True)
                if isinstance(data, list):
                    data = class_type.from_list(data)
                else:
                    data = class_type.from_dict(data, input=True)
            elif isinstance(data, dict):
                # Data directly provided as dict
                data = class_type.from_dict(data, input=True)
            elif isinstance(data, list) and all(
                isinstance(element, dict) for element in data
            ):
                data = class_type.from_list(data)
            if data:
                # pylint: disable=protected-access
                if not isinstance(data, list):
                    kwargs["data"] = data._get_api_params()
                else:
                    kwargs["data"] = [element._get_api_params() for element in data]
            else:
                raise ValueError("Please provide valid data or path to valid data")
            return func(self, *args, **kwargs)

        return wrapper

    return wrap


def request_ok(func):
    """Wrapper that checks if http request is valid (raise for status is called)

    Parameters
    ----------
    func : function
        Function
    """

    def wrapper(self, *args, **kwargs):
        request = func(self, *args, **kwargs)
        try:
            request.raise_for_status()
            return request
        except Exception:
            raise ValueError(request.text)

    return wrapper


def with_token(func):
    def wrapper(self, *args, **kwargs):
        # Update the session headers with a valid token
        access_token = self.get_valid_access_token()
        self.session.headers.update(
            {
                "Content-Type": "application/json",
                "Authorization": f"Bearer {access_token}",
            }
        )
        return func(self, *args, **kwargs)

    return wrapper



import datetime
from pathlib import Path
from typing import Union

from whiffle_client.forecast.components.observation_model import ObservationModel
from whiffle_client.decorators import load_data
from whiffle_client.common.mapping.base import BaseMapping


class ObservationModelEndpoints(BaseMapping):
    """Observation model endpoint

    Parameters
    ----------
    BaseMapping : Basemapping
        Base mapping
    """

    URL = "/api/v1/observations"
    RESOURCE_TYPE = "ObservationModel"

    # Observation models commands
    @load_data(ObservationModel)
    def add(
        self,
        data: Union[str, dict, Path, ObservationModel] = None,
    ) -> ObservationModel:
        """Add new Observation Model

        Parameters
        ----------
        data : Union[str, dict, Path, ObservationModel], optional
            Either a path to a yaml, the data itself as a dictionary or a ObservationModel object
            containing the parameters that define the observation model specs, by default None

        Returns
        -------
        ObservationModel
            Object instance of the observation model
        """
        request = self.session.post_request(
            f"{self.session.server_url}{self.URL}", data=data
        )
        return ObservationModel.from_dict(request.json())

    def get(self):
        """Get method for observation

        Raises
        ------
        NotImplementedError
            Not implemented; use get_all method
        """
        raise NotImplementedError(
            "Get method for observations not implemented; use get_all method."
        )

    # pylint: disable=arguments-differ
    def get_all(
        self,
        asset_name: str,
        quantity: str,
        time_after: datetime.datetime,
        time_before: datetime.datetime,
    ) -> list[ObservationModel]:
        """Get a list of all the Observation Models available to the user in predefined period

        Returns
        -------
        list[ObservationModel]
            List of ObservationModel object instances
        """
        observations = []
        data = self.session.get_request(
            f"{self.session.server_url}{self.URL}",
            params={
                "asset_name": asset_name,
                "quantity": quantity,
                "time_after": time_after.isoformat(),
                "time_before": time_before.isoformat(),
            },
        ).json()
        observations.extend(
            ObservationModel.from_dict(observation_model)
            for observation_model in data.get("results", [])
        )

        while isinstance(data.get("next"), str):
            data = self.session.get_request(data["next"]).json()
            observations.extend(
                ObservationModel.from_dict(observation_model)
                for observation_model in data.get("results", [])
            )
        return observations

    def edit(self):
        """Abstract method for edit

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def delete(self):
        """Abstract method for delete

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError

    def download(self):
        """Abstract method for download

        Raises
        ------
        NotImplementedError
            Not implemented
        """
        raise NotImplementedError
